""" Tests for the workflow service """

import json
import logging
import os
import pathlib
import shutil
import sys
from typing import List
# pylint: disable=E0401, R0201, W0212
from unittest.mock import MagicMock, patch

import pytest

from workspaces.workflow.schema import Workflow, WorkflowRequest, WorkflowRequestFile
from workspaces.workflow.services.workflow_info import WorkflowInfo
from workspaces.workflow.services.workflow_service import WorkflowService

pytest_plugins = ["testing.utils.conftest"]

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

"""
Tests for WorkflowService
"""


files = [
    WorkflowRequestFile(
        workflow_request_id=-1,
        filename="null.condor",
        content=b"executable = null\n"
        + b"arguments = {{arguments}}\n"
        + b"error = null.err\n"
        + b"log = condor.log\n\n\n"
        + b"queue",
    ),
    WorkflowRequestFile(
        workflow_request_id=-1,
        filename="null.dag",
        content=b"JOB null null.condor",
    ),
    WorkflowRequestFile(
        workflow_request_id=-1,
        filename="null.sh",
        content=b"#!/bin/sh\n\n" + b"null $*",
    ),
]

def test_do_nothing():
    """
    (make test will croak if there isn't at least one test)
    :return:
    """
    pass

@pytest.mark.usefixtures("mock_workflow_info", "mock_workflow_service", "test_router")
class TestWorkflowService:
    """Our test class"""

    requests = []

    def test_create_workflow_request(self, mock_workflow_info: WorkflowInfo):
        """
        Is a workflow request created as expected from workflow info?

        :param mock_workflow_info:
        :return:
        """
        null_workflow = mock_workflow_info.lookup_workflow_definition("null")
        assert null_workflow is not None

        request = mock_workflow_info.create_workflow_request(null_workflow,
            json.loads('{"argument":"-g"}'),
        )
        self.requests.append(request)
        assert request in self.requests

    def test_attach_file_to_request(
        self,
        mock_workflow_info: WorkflowInfo,
        mock_workflow_requests: List[WorkflowRequest],
    ):
        """
        Is a file attached to workflow requests as expected?

        :param mock_workflow_info:
        :param mock_workflow_requests:
        :return:
        """
        request = mock_workflow_requests[0]
        test_filename = "test.txt"
        test_content = b"testing file additions"
        file = mock_workflow_info.save_file(request=request, filename=test_filename, content=test_content)
        mock_workflow_info.session.add.assert_called_with(file)
        assert mock_workflow_info.session.flush.call_count == 1

    @patch("pathlib.Path.write_bytes")
    @patch("os.chown")
    @patch("os.chmod")
    @pytest.mark.skip("breaks the CI?")
    def test_prepare_files_for_condor(
        self,
        mock_chmod,
        mock_chown,
        mock_write,
        mock_workflow_service: WorkflowService,
        mock_workflow_requests: List[WorkflowRequest],
    ):
        """
        Make sure files are prepared for Condor.

        :param mock_workflow_service:
        :return:
        """
        with patch("os.mkdir") as make_dir:
            temp = mock_workflow_service._make_temp_directory(mock_workflow_requests[0])
            mock_workflow_service._prepare_files_for_condor(files, temp)
            assert mock_write.call_count == 3


    @patch("workspaces.workflow.schema.Workflow.render_templates")
    @patch("messaging.messenger.MessageSender.send_message")
    @patch("json.loads")
    @patch("subprocess.Popen")
    @pytest.mark.skip("breaks the CI?")
    def test_render_with_metadata(
        self,
        mock_subprocess,
        mock_loads,
        mock_router,
        mock_render,
        mock_workflow_service: WorkflowService,
        mock_workflow_requests: List[WorkflowRequest],
    ):
        definition = Workflow(workflow_name="std_calibration")
        request = mock_workflow_requests[1]
        tempdir = pathlib.Path("/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmp1234abcd")
        mock_workflow_service._render_with_metadata(request, tempdir, definition)

        assert mock_subprocess.call_count == 1
        assert mock_loads.call_count == 1
        assert mock_router.call_count == 0
        assert mock_render.call_count == 1

    @patch("workspaces.workflow.services.workflow_info.WorkflowInfo.save_file")
    @pytest.mark.skip("breaks the CI?")
    def test_determine_usable_files(
        self,
        mock_save,
        mock_workflow_requests: List[WorkflowRequest],
        mock_workflow_service: WorkflowService,
    ):
        request = mock_workflow_requests[0]
        usable = mock_workflow_service._determine_usable_files(request, files)
        assert len(usable) == 3

    @pytest.mark.skip("breaks the CI?")
    def test_execute_prepared(self, mock_workflow_service: WorkflowService, tmp_path: pathlib.Path):
        # Create fake folder and DAG file
        fake_folder = tmp_path / "test"
        fake_folder.mkdir()
        fake_dagfile = fake_folder / "test.dag"
        fake_dagfile.touch()

        orig_submit_dag = mock_workflow_service._submit_dag
        mock_workflow_service._submit_dag = MagicMock()

        # Test DAG submission
        mock_workflow_service._execute_prepared(fake_folder, False)
        mock_workflow_service._submit_dag.assert_called_with(fake_folder, fake_dagfile, False)

        # DAG cleanup
        mock_workflow_service._submit_dag = orig_submit_dag
        os.remove(fake_dagfile)

        # Create fake submit file
        fake_jobfile = fake_folder / "test.condor"
        fake_jobfile.touch()

        orig_submit_job = mock_workflow_service._submit_job
        mock_workflow_service._submit_job = MagicMock()

        # Test job submission
        mock_workflow_service._execute_prepared(fake_folder, False)
        mock_workflow_service._submit_job.assert_called_with(fake_folder, fake_jobfile, False)

        # Cleanup
        mock_workflow_service._submit_job = orig_submit_job
        shutil.rmtree(tmp_path)

    @pytest.mark.skip("breaks the CI?")
    def test_submit_job(self, mock_workflow_service: WorkflowService):
        fake_folder = pathlib.Path("fake_folder")
        fake_jobfile = pathlib.Path("fake.submit")
        fake_jobfile.touch()
        with patch("workspaces.workflow.services.workflow_service.subprocess.run") as mock_sp_run:
            log_path = mock_workflow_service._submit_job(fake_folder, fake_jobfile, False)
            mock_sp_run.assert_called_with(
                ["condor_submit", fake_jobfile.name],
                cwd=str(fake_folder.absolute()),
                preexec_fn=mock_workflow_service._switch_to_submituser(False),
            )
            assert log_path.name == "fake.log"

        os.remove(fake_jobfile)

    @pytest.mark.skip("breaks the CI?")
    def test_submit_dag(self, mock_workflow_service: WorkflowService):
        fake_folder = pathlib.Path("fake_folder")
        fake_dagfile = pathlib.Path("fake.dag")
        with patch("workspaces.workflow.services.workflow_service.subprocess.run") as mock_sp_run:
            log_path = mock_workflow_service._submit_dag(fake_folder, fake_dagfile, False)
            mock_sp_run.assert_called_with(
                ["condor_submit_dag", "-f", fake_dagfile.name],
                cwd=str(fake_folder.absolute()),
                preexec_fn=mock_workflow_service._switch_to_submituser(False),
            )
            assert log_path.name == "fake.dag.dagman.log"

    @pytest.mark.skip("breaks the CI?")
    def test_get_jobfile_log(self, tmp_path: pathlib.Path, mock_workflow_service: WorkflowService):
        with open(tmp_path / "fake.submit", "w") as fake_jobfile:
            # Test with file that doesn't containe "log = " entry
            log_path = mock_workflow_service._get_job_logfile_name(tmp_path / "fake.submit")
            assert log_path.name == "fake.log"
            # Write "log = " entry
            fake_jobfile.write("log = test.log")

        # Test with same file that now contains "log = " entry
        log_path = mock_workflow_service._get_job_logfile_name(tmp_path / "fake.submit")
        assert log_path.name == "test.log"

        # Remove tmp_path just in case
        shutil.rmtree(tmp_path)

    @pytest.mark.skip("breaks the CI?")
    def test_get_dagfile_log(self, mock_workflow_service: WorkflowService):
        log_path = mock_workflow_service._get_dag_logfile_name(pathlib.Path("fake.dag"))
        assert log_path.name == "fake.dag.dagman.log"
